from __future__ import absolute_import, division, print_function
import numpy as np
from NuRadioReco.utilities import units
from NuRadioMC.utilities import cross_sections
from radiotools import helper as hp
import logging
logger = logging.getLogger("utilities.earth_attenuation")

AMU = 1.66e-27 * units.kg


def get_weight(theta_nu, pnu, flavors, mode='simple', cross_section_type='ctw',
               vertex_position=None, phi_nu=None):
    """
    calculates neutrino weight due to Earth absorption for different models

    Parameters
    ----------
    theta_nu: float or array of floats
        the zenith angle of the neutrino direction (where it came from, i.e., opposite to the direction of propagation)
    pnu: float or array of floats
        the momentum of the neutrino
    mode: string
        * 'simple': assuming interaction happens at the surface and approximating the Earth with constant density
        * 'core_mantle_crust_simple': assuming interaction happens at the surface and approximating the Earth with 3 layers of constant density
        * 'core_mantle_crust': approximating the Earth with 3 layers of constant density, path through Earth to interaction vertex is considered
        * 'PREM': density of Earth is parameterized as a fuction of radius, path through Earth to interaction vertex is considered
    cross_section_type: string
        'ghandi', 'ctw' or 'csms' (see description in `cross_sections.py`)
    vertex_position: 3-dim array or None (default)
        the position of the neutrino interaction
    phi_nu: float
        the azimuth angle of the neutrino direction
    """
    if(mode == 'simple'):
        return get_simple_weight(theta_nu, pnu, cross_section_type=cross_section_type)
    elif (mode == "core_mantle_crust_simple"):
        return get_core_mantle_crust_weight(theta_nu, pnu, flavors, cross_section_type=cross_section_type)
    elif (mode == "core_mantle_crust"):
        earth = CoreMantleCrustModel()
        direction = hp.spherical_to_cartesian(theta_nu, phi_nu)
        slant_depth = earth.slant_depth(vertex_position, direction)
        # by requesting the interaction length for a density of 1, we get it in units of length**2/weight
        L_int = cross_sections.get_interaction_length(pnu, density=1., flavor=flavors, inttype='total',
                                                      cross_section_type=cross_section_type)
        return np.exp(-slant_depth / L_int)
    elif (mode == "PREM"):
        earth = PREM()
        direction = hp.spherical_to_cartesian(theta_nu, phi_nu)
        slant_depth = earth.slant_depth(vertex_position, direction)
        # by requesting the interaction length for a density of 1, we get it in units of length**2/weight
        L_int = cross_sections.get_interaction_length(pnu, density=1., flavor=flavors, inttype='total',
                                                      cross_section_type=cross_section_type)
        return np.exp(-slant_depth / L_int)
    elif (mode == "None"):
        return 1.
    else:
        logger.error('mode {} not supported'.format(mode))
        raise NotImplementedError


def get_simple_weight(theta_nu, pnu, cross_section_type='ctw'):
    """
    calculates neutrino weight due to Earth absorption, i.e. probability of the
    neutrino to reach the detector

    simple parametrization using only momentum and zenith angle information
    of the neutrino, adapted from ShelfMC

    Parameters
    ----------
    theta_nu: float or array of floats
        the zenith angle of the neutrino direction (where it came from, i.e., opposite to the direction of propagation)
    pnu: float or array of floats
        the momentum of the neutrino
    """
    R_earth = 6357390 * units.m
    DensityCRUST = 2900 * units.kg / units.m ** 3
    if(theta_nu <= 0.5 * np.pi):  # coming from above
        return np.ones_like(theta_nu)
    else:  # coming from below
        sigma = cross_sections.get_nu_cross_section(pnu, flavors=0, cross_section_type=cross_section_type)
        d = -2 * R_earth * np.cos(theta_nu)
        return np.exp(-d * sigma * DensityCRUST / AMU)


def get_core_mantle_crust_weight(theta_nu, pnu, flavors, cross_section_type='ctw'):
    """
    calculates neutrino weight due to Earth absorption with a three layers earth model, i.e. probability of the
    neutrino to reach the detector

    simple parametrization using momentum, zenith angle, flavor and current type information
    of the neutrino

    parameters from ARAsim (2018)

    Parameters
    ----------
    theta_nu: float or array of floats
        the zenith angle of the neutrino direction (where it came from, i.e., opposite to the direction of propagation)
    pnu: float or array of floats
        the momentum of the neutrino
    flavors: float or array of floats
        the flavor of the neutrino
    """
    R_EARTH = 6.378140e6 * units.m
    densities = np.array([14000.0, 3400.0, 2900.0]) * units.kg / units.m ** 3  # inner layer, middle layer, outer layer
    radii = np.array([3.46e6 * units.m, R_EARTH - 4.0e4 * units.m, R_EARTH])  # average radii of boundaries between earth layers
    sigma = cross_sections.get_nu_cross_section(pnu, flavors, cross_section_type=cross_section_type)
    if(theta_nu <= 0.5 * np.pi):  # coming from above
        return np.ones_like(theta_nu)
    elif (theta_nu <= np.pi - np.arcsin(radii[1] / radii[2])):  # only go through the outer layer
        d_outer = -2 * R_EARTH * np.cos(theta_nu)
        weight = np.exp(-d_outer * sigma * densities[2] / AMU)
    elif (theta_nu <= np.pi - np.arcsin(radii[0] / radii[2])):  # only go through the outer and middle layer
        d_middle = 2 * np.sqrt(radii[1] * radii[1] - radii[2] * radii[2] * np.sin(np.pi - theta_nu) * np.sin(np.pi - theta_nu))
        d_outer = -2 * R_EARTH * np.cos(theta_nu) - d_middle
        weight = np.exp(-d_outer * sigma * densities[2] / AMU - d_middle * sigma * densities[1] / AMU)
    else:  # go through all three layers
        d_inner = 2 * np.sqrt(radii[0] * radii[0] - radii[2] * radii[2] * np.sin(np.pi - theta_nu) * np.sin(np.pi - theta_nu))
        d_middle = 2 * np.sqrt(radii[1] * radii[1] - radii[2] * radii[2] * np.sin(np.pi - theta_nu) * np.sin(np.pi - theta_nu)) - d_inner
        d_outer = -2 * R_EARTH * np.cos(theta_nu) - d_middle - d_inner
        weight = np.exp(-d_outer * sigma * densities[2] / AMU - d_middle * sigma * densities[1] / AMU - d_inner * sigma * densities[0] / AMU)
    return weight


# PREM class from pyrex: https://github.com/bhokansonfasig/pyrex/blob/d84a3270efa19fb4a21590510f7c3458845c9600/pyrex/earth_model.py
class PREM:
    """
    Class describing the Earth's density.

    Uses densities from the Preliminary reference Earth Model (PREM).

    Attributes
    ----------
    earth_radius : float
        Mean radius of the Earth (m).
    radii : tuple
        Boundary radii at which the functional form of the density of the
        Earth changes. The density function in `densities` at index `i`
        corresponds to the radius range from radius at index `i-1` to radius
        at index `i`.
    densities : tuple
        Functions which calculate the density of the Earth in a
        specific radius range as described by `radii`. The parameter of each
        function is the fractional radius, e.g. radius divided by
        `earth_radius`. Scalar values denote constant density over the range of
        radii.

    Notes
    -----
    The density calculation is based on the Preliminary reference Earth Model
    [1]_.

    References
    ----------
    .. [1] A. Dziewonski & D. Anderson, "Preliminary reference Earth model."
        Physics of the Earth and Planetary Interiors **25**, 297–356 (1981). :doi:`10.1016/0031-9201(81)90046-7`

    """
    earth_radius = 6.3710e6 * units.m

    radii = (1.2215e6 * units.m, 3.4800e6 * units.m, 5.7010e6 * units.m, 5.7710e6 * units.m, 5.9710e6 * units.m,
             6.1510e6 * units.m, 6.3466e6 * units.m, 6.3560e6 * units.m, 6.3680e6 * units.m, earth_radius)

    # `x` is fraction of earth radius
    densities = (
        lambda x: (13.0885 * units.g / units.cm ** 3 - 8.8381 * units.g / units.cm ** 3 * x ** 2),
        lambda x: 12.5815 * units.g / units.cm ** 3 - 1.2638 * units.g / units.cm ** 3 * x - 3.6426 * units.g / units.cm ** 3 * x ** 2 - 5.5281 * units.g / units.cm ** 3 * x ** 3,
        lambda x: 7.9565 * units.g / units.cm ** 3 - 6.4761 * units.g / units.cm ** 3 * x + 5.5283 * units.g / units.cm ** 3 * x ** 2 - 3.0807 * units.g / units.cm ** 3 * x ** 3,
        lambda x: 5.3197 * units.g / units.cm ** 3 - 1.4836 * units.g / units.cm ** 3 * x,
        lambda x: 11.2494 * units.g / units.cm ** 3 - 8.0298 * units.g / units.cm ** 3 * x,
        lambda x: 7.1089 * units.g / units.cm ** 3 - 3.8045 * units.g / units.cm ** 3 * x,
        lambda x: 2.691 * units.g / units.cm ** 3 + 0.6924 * units.g / units.cm ** 3 * x,
        2.9 * units.g / units.cm ** 3,
        2.6 * units.g / units.cm ** 3,
        1.02 * units.g / units.cm ** 3
    )

    def density(self, r):
        """
        Calculates the Earth's density at a given radius.

        Supports passing an array of radii or a single radius.

        Parameters
        ----------
        r : array_like
            Radius (m) at which to calculate density.

        Returns
        -------
        array_like
            Density (g/cm^3) of the Earth at the given radii.

        """
        r = np.array(r)
        radius_bounds = np.concatenate(([0], self.radii))
        conditions = list((lower <= r) & (r < upper) for lower, upper in
                          zip(radius_bounds[:-1], radius_bounds[1:]))
        return np.piecewise(r / self.earth_radius, conditions, self.densities)

    def slant_depth(self, endpoint, direction, step=500 * units.m):
        """
        Calculates the column density of a chord cutting through Earth.

        Integrates the Earth's density along the chord, resulting in a column
        density (or material thickness) with units of mass per area.

        Parameters
        ----------
        endpoint : array_like
            Vector position of the chord endpoint, in a coordinate system
            centered on the surface of the Earth (e.g. a negative third
            coordinate represents the depth below the surface).
        direction : array_like
            Vector direction of the chord, in a coordinate system
            centered on the surface of the Earth (e.g. a negative third
            coordinate represents the chord pointing into the Earth).
        step : float, optional
            Step size for the integration.

        Returns
        -------
        float
            Column density along the chord starting from `depth` and
            passing through the Earth at `angle`.

        See Also
        --------
        PREM.density : Calculates the Earth's density at a given radius.

        """
        # Convert to Earth-centric coordiante system (e.g. center of the Earth
        # is at (0, 0, 0))
        endpoint = np.array([endpoint[0], endpoint[1],
                             endpoint[2] + self.earth_radius])
        direction /= np.linalg.norm(direction)
        dot_prod = np.dot(endpoint, direction)
        # Check for intersection of line and sphere
        discriminant = dot_prod ** 2 - np.sum(endpoint ** 2) + self.earth_radius ** 2
        if discriminant <= 0:
            return 0
        # Calculate the distance at which the line intersects the sphere
        distance = -dot_prod + np.sqrt(discriminant)
        if distance <= 0:
            return 0
        # Parameterize line integral with ts from 0 to 1, with steps just under
        # the given step size (in meters)
        n_steps = int(distance / step)
        if distance % step:
            n_steps += 1
        ts = np.linspace(0, 1, n_steps)
        xs = endpoint[0] + ts * distance * direction[0]
        ys = endpoint[1] + ts * distance * direction[1]
        zs = endpoint[2] + ts * distance * direction[2]
        rs = np.sqrt(xs ** 2 + ys ** 2 + zs ** 2)
        rhos = self.density(rs)
        return np.trapz(rhos * distance, ts)


class CoreMantleCrustModel(PREM):
    """
    Class describing the Earth's density.

    Uses densities from the Core-Mantle-Crust model. Parameters from ARASim

    Attributes
    ----------
    earth_radius : float
        Mean radius of the Earth (m).
    radii : tuple
        Boundary radii (m) at which the functional form of the density of the
        Earth changes. The density function in `densities` at index `i`
        corresponds to the radius range from radius at index `i-1` to radius
        at index `i`.
    densities : tuple
        Functions which calculate the density of the Earth (g/cm^3) in a
        specific radius range as described by `radii`. The parameter of each
        function is the fractional radius, e.g. radius divided by
        `earth_radius`. Scalar values denote constant density over the range of
        radii.

    """
    earth_radius = 6.378140e6 * units.m

    radii = (np.sqrt(1.2e13) * units.m, earth_radius - 4e4 * units.m, earth_radius)

    densities = (14 * units.g / units.cm ** 3, 3.4 * units.g / units.cm ** 3, 2.9 * units.g / units.cm ** 3)




#!/usr/bin/env python
from NuRadioMC.SignalGen.askaryan import get_time_trace, get_frequency_spectrum
from NuRadioReco.utilities import units
from NuRadioReco.utilities import io_utilities
import numpy as np
from numpy import testing
import sys

try:
    reference_file = sys.argv[1]
except:
    reference_file = "reference_v1.pkl"

print('Using reference file {}'.format(reference_file))

np.random.seed(0)

n_index = 1.78
domega = 0.05 * units.deg
theta = np.arccos(1. / n_index) + domega

dt = 0.5 * units.ns
n_samples = 256
R = 1 * units.km

models = ['Alvarez2009', 'ARZ2019', 'Alvarez2000', 'ARZ2020']
shower_types = ['EM', 'HAD']

Es = 10 ** np.linspace(15, 19, 5) * units.eV
domegas = np.linspace(-5, 5, 10) * units.deg
thetas = np.arccos(1. / n_index) + domegas

reference = io_utilities.read_pickle(reference_file, encoding='latin1')
i = -1
for model in models:
    for E in Es:
        for shower_type in shower_types:
            for theta in thetas:
                i += 1
                trace = get_time_trace(E, theta, n_samples, dt, shower_type, n_index, R, model, seed=1234)
                try:
                    testing.assert_almost_equal(trace, reference[i])
                except AssertionError as e:
                    print(f"error in model {model}, shower type {shower_type} theta = {theta/units.deg:.2f}deg")
                    print(e)
                    raise(e)

print('SignalGen test passed without any issues!')



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function
import argparse
# import detector simulation modules
import NuRadioReco.modules.efieldToVoltageConverter
import NuRadioReco.modules.trigger.highLowThreshold
import NuRadioReco.modules.trigger.simpleThreshold
import NuRadioReco.modules.channelResampler
import NuRadioReco.modules.channelBandPassFilter
import NuRadioReco.modules.triggerTimeAdjuster
import NuRadioReco.modules.channelGenericNoiseAdder
import NuRadioReco.modules.electricFieldResampler
from NuRadioReco.utilities import units
from NuRadioMC.simulation import simulation
import logging
import numpy as np
np.random.seed(0)
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("runstrawman")

# initialize detector sim modules
efieldToVoltageConverter = NuRadioReco.modules.efieldToVoltageConverter.efieldToVoltageConverter()
efieldToVoltageConverter.begin()
triggerSimulatorHighLow = NuRadioReco.modules.trigger.highLowThreshold.triggerSimulator()
triggerSimulatorSimple = NuRadioReco.modules.trigger.simpleThreshold.triggerSimulator()
channelBandPassFilter = NuRadioReco.modules.channelBandPassFilter.channelBandPassFilter()
channelGenericNoiseAdder = NuRadioReco.modules.channelGenericNoiseAdder.channelGenericNoiseAdder()
channelResampler = NuRadioReco.modules.channelResampler.channelResampler()
electricFieldResampler = NuRadioReco.modules.electricFieldResampler.electricFieldResampler()
triggerTimeAdjuster = NuRadioReco.modules.triggerTimeAdjuster.triggerTimeAdjuster()


class mySimulation(simulation.simulation):

    def _detector_simulation(self):

        efieldToVoltageConverter.run(self._evt, self._station, self._det)  # convolve efield with antenna pattern
        # downsample trace to internal simulation sampling rate (the efieldToVoltageConverter upsamples the trace to
        # 20 GHz by default to achive a good time resolution when the two signals from the two signal paths are added)
        channelResampler.run(self._evt, self._station, self._det, sampling_rate=1. / self._dt)

        if self._is_simulate_noise():
            max_freq = 0.5 / self._dt
            norm = self._get_noise_normalization(self._station.get_id())  # assuming the same noise level for all stations
            Vrms = self._Vrms / (norm / (max_freq)) ** 0.5  # normalize noise level to the bandwidth its generated for
            channelGenericNoiseAdder.run(self._evt, self._station, self._det, amplitude=Vrms, min_freq=0 * units.MHz,
                                         max_freq=max_freq, type='rayleigh')

        # bandpass filter trace, the upper bound is higher then the sampling rate which makes it just a highpass filter
        channelBandPassFilter.run(self._evt, self._station, self._det, passband=[80 * units.MHz, 1000 * units.GHz],
                                  filter_type='butter', order=2)
        channelBandPassFilter.run(self._evt, self._station, self._det, passband=[0, 500 * units.MHz],
                                  filter_type='butter', order=10)
        # first run a simple threshold trigger
        triggerSimulatorSimple.run(self._evt, self._station, self._det,
                             threshold=3 * self._Vrms,
                             triggered_channels=None,  # run trigger on all channels
                             number_concidences=1,
                             trigger_name='simple_threshold')  # the name of the trigger

        # run a high/low trigger on the 4 downward pointing LPDAs
        triggerSimulatorHighLow.run(self._evt, self._station, self._det,
                                    threshold_high=4 * self._Vrms,
                                    threshold_low=-4 * self._Vrms,
                                    triggered_channels=[0, 1, 2, 3],  # select the LPDA channels
                                    number_concidences=2,  # 2/4 majority logic
                                    trigger_name='LPDA_2of4_4.1sigma',
                                    set_not_triggered=(not self._station.has_triggered("simple_threshold")))  # calculate more time consuming ARIANNA trigger only if station passes simple trigger

        # run a high/low trigger on the 4 surface dipoles
        triggerSimulatorHighLow.run(self._evt, self._station, self._det,
                                    threshold_high=3 * self._Vrms,
                                    threshold_low=-3 * self._Vrms,
                                    triggered_channels=[4, 5, 6, 7],  # select the bicone channels
                                    number_concidences=4,  # 4/4 majority logic
                                    trigger_name='surface_dipoles_4of4_3sigma',
                                    set_not_triggered=(not self._station.has_triggered("simple_threshold")))  # calculate more time consuming ARIANNA trigger only if station passes simple trigger
        triggerTimeAdjuster.run(self._evt, self._station, self._det)


parser = argparse.ArgumentParser(description='Run NuRadioMC simulation')
parser.add_argument('inputfilename', type=str,
                    help='path to NuRadioMC input event list')
parser.add_argument('detectordescription', type=str,
                    help='path to file containing the detector description')
parser.add_argument('config', type=str,
                    help='NuRadioMC yaml config file')
parser.add_argument('outputfilename', type=str,
                    help='hdf5 output filename')
parser.add_argument('outputfilenameNuRadioReco', type=str, nargs='?', default=None,
                    help='outputfilename of NuRadioReco detector sim file')
args = parser.parse_args()

sim = mySimulation(inputfilename=args.inputfilename,
                            outputfilename=args.outputfilename,
                            detectorfile=args.detectordescription,
                            outputfilenameNuRadioReco=args.outputfilenameNuRadioReco,
                            config_file=args.config,
                            write_mode='mini',
                            default_detector_station=101,
                            file_overwrite=True)
sim.run()




from __future__ import absolute_import, division, print_function
import numpy as np
from NuRadioReco.utilities import units
from six import iterkeys, iteritems
from scipy import constants
from scipy.integrate import quad
from scipy.interpolate import interp1d
from scipy.optimize import fsolve
import h5py
from NuRadioMC.EvtGen.generator import write_events_to_hdf5
import logging
logger = logging.getLogger("EventGen")
logging.basicConfig()

VERSION_MAJOR = 1
VERSION_MINOR = 1


def generate_my_events(filename, n_events):
    """
    Event generator skeleton

    Parameters
    ----------
    filename: string
        the output filename of the hdf5 file
    n_events: int
        number of events to generate
    """

    # first set the meta attributes
    attributes = {}
    n_events = int(n_events)
    attributes['n_events'] = n_events
    attributes['start_event_id'] = 0
#     attributes['fiducial_rmin'] = 0
#     attributes['fiducial_rmax'] = 1 * units.km
#     attributes['fiducial_zmin'] = 0 * units.m
#     attributes['fiducial_zmax'] = -2 * units.km
#     attributes['rmin'] = 0
#     attributes['rmax'] = 1 * units.km
#     attributes['zmin'] = 0 * units.m
#     attributes['zmax'] = -2 * units.km
#     attributes['Emin'] = 1 * units.EeV
#     attributes['Emax'] = 1 * units.EeV

    # now generate the events and fill all required data sets
    # here we fill all data sets with dummy values
    data_sets = {}
    # the 'neutrino direction' needs to be set but are irrelevant for the simulation, because we simulate a
    # uniform emitter
    data_sets["azimuths"] = np.ones(n_events)
    data_sets["zeniths"] = np.ones(n_events)

    # define the emitter positions. X/Y are the easting/northing coordinates of the SPICE core
    data_sets["xx"] = np.ones(n_events) * 42600 * units.feet
    data_sets["yy"] = np.ones(n_events) * 48800 * units.feet
    # simualte different depth
    data_sets["zz"] = -np.linspace(0, 1800, n_events) * units.m
    data_sets["event_ids"] = np.arange(n_events)
    data_sets["n_interaction"] = np.ones(n_events, dtype=np.int)
    data_sets["interaction_type"] = np.array(['had'] * n_events)

    # again these parameters are irrelevant for our simulation but still need to be set
    data_sets["flavors"] = np.array([12 for i in range(n_events)])
    data_sets["energies"] = np.ones(n_events) * 1 * units.eV
    data_sets["inelasticity"] = np.ones(n_events) * 0.5

    # write events to file
    write_events_to_hdf5(filename, data_sets, attributes)


if __name__ == "__main__":
    generate_my_events("input_spice.hdf5", 1000)



import matplotlib.pyplot as plt
import numpy as np
import time
from NuRadioMC.SignalProp import analyticraytracing as ray
from NuRadioMC.utilities import medium
from NuRadioReco.utilities import units
import logging
from numpy import testing
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('test_raytracing')

ice = medium.southpole_simple()

np.random.seed(0)  # set seed to have reproducible results
n_events = int(1e3)
rmin = 50. * units.m
rmax = 3. * units.km
zmin = 0. * units.m
zmax = -3. * units.km
rr = np.random.triangular(rmin, rmax, rmax, n_events)
phiphi = np.random.uniform(0, 2 * np.pi, n_events)
xx = rr * np.cos(phiphi)
yy = rr * np.sin(phiphi)
zz = np.random.uniform(zmin, zmax, n_events)

points = np.array([xx, yy, zz]).T
x_receiver = np.array([0., 0., -5.])

results_C0s_cpp = np.zeros((n_events, 2))
n_freqs = 256//2 + 1
# n_freqs = 5
results_A_cpp = np.zeros((n_events, 2, n_freqs))
t_start = time.time()
ff = np.linspace(0, 500*units.MHz, n_freqs)
# tt = 0
for iX, x in enumerate(points):
#     t_start2 = time.time()
    r = ray.ray_tracing(x, x_receiver, ice)
#     tt += (time.time() - t_start2)
    r.find_solutions()
    if(r.has_solution()):
        for iS in range(r.get_number_of_solutions()):
            results_C0s_cpp[iX, iS] = r.get_results()[iS]['C0']
            results_A_cpp[iX, iS] = r.get_attenuation(iS, ff)
t_cpp = time.time() - t_start
print("CPP time = {:.1f} seconds = {:.2f}ms/event".format(t_cpp, 1000. * t_cpp / n_events))
# print("CPP time = {:.1f} seconds = {:.2f}ms/event".format(tt, 1000. * tt / n_events))


results_C0s_python = np.zeros((n_events, 2))
results_A_python = np.zeros((n_events, 2, n_freqs))
ray.cpp_available = False
t_start = time.time()
for iX, x in enumerate(points):
    r = ray.ray_tracing(x, x_receiver, ice)
    r.find_solutions()
    if(r.has_solution()):
        for iS in range(r.get_number_of_solutions()):
            results_C0s_python[iX, iS] = r.get_results()[iS]['C0']
            results_A_python[iX, iS] = r.get_attenuation(iS, ff)
t_python = time.time() - t_start
print("Python time = {:.1f} seconds = {:.2f}ms/event".format(t_python, 1000. * t_python / n_events))

testing.assert_allclose(results_C0s_cpp, results_C0s_python, atol=1e-08, rtol=1e-05)
testing.assert_allclose(results_A_cpp, results_A_python, rtol=1e-2, atol=1e-3)

print('T01test_python_vs_cpp passed without issues')



import numpy as np
import h5py
from scipy import interpolate
import glob
from six import iteritems
import json
import os
import copy

from NuRadioReco.utilities import units
from NuRadioMC.EvtGen.generator import get_projected_area_cylinder, get_projected_area_cylinder_integral

import logging
logger = logging.getLogger("Veff")
logging.basicConfig()

# collection of utility function regarding the calculation of the effective volume of a neutrino detector


def get_triggered(fin):
    """
    Computes an array indicating the triggered events.
    If a double bang is seen, removes the second bang from the actual triggered
    array so as not to count twice the same event for the effective volume.

    Parameters
    ----------
    fin: dictionary
       Dictionary containing the output data sets from the simulation

    Returns
    -------
    triggered: numpy array with bools
       The bools indicate if the events have triggered
    """

    triggered = np.copy(fin['triggered'])

    if (len(triggered) == 0):
        return triggered

    mask_secondaries = np.array(fin['n_interaction']) > 1
    if (True not in mask_secondaries):
        return triggered

    # We count the multiple triggering bangs as a single triggered event
    for event_id in np.unique(np.array(fin['event_ids'])[mask_secondaries]):
        mask_interactions = np.array(fin['event_ids']) == event_id
        multiple_interaction_indexes = np.squeeze(np.argwhere(np.array(fin['event_ids']) == event_id))
        if (len(multiple_interaction_indexes) == 1):
            continue

        for int_index in multiple_interaction_indexes[1:]:
            triggered[int_index] = False
        triggered[multiple_interaction_indexes[0]] = True

    return triggered


def get_Aeff_proposal(folder, trigger_combinations={}, station=101):
    """
    Calculates the effective area from NuRadioMC hdf5 files simulated after
    using PROPOSAL as the lepton propagator. The interaction length is already
    factorised thanks to the energy losses returned by PROPOSAL, which in turn
    can trigger our radio array.

    Parameters
    ----------
    folder: string
        folder conaining the hdf5 files, one per energy
    trigger_combinations: dict, optional
        keys are the names of triggers to calculate. Values are dicts again:
            * 'triggers': list of strings
                name of individual triggers that are combines with an OR
            the following additional options are optional
            * 'efficiency': string
                the signal efficiency vs. SNR (=Vmax/Vrms) to use. E.g. 'Chris'
            * 'efficiency_scale': float
                rescaling of the efficiency curve by SNR' = SNR * scale
            * 'n_reflections': int
                the number of bottom reflections of the ray tracing solution that likely triggered
                assuming that the solution with the shortest travel time caused the trigger, only considering channel 0

    station: int
        the station that should be considered

    Returns
    ----------
    list of dictionary. Each file is one entry. The dictionary keys store all relevant properties
    """
    Aeff_output = []
    trigger_names = None
    trigger_names_dict = {}
    prev_deposited = None
    deposited = False

    if(len(glob.glob(os.path.join(folder, '*.hdf5'))) == 0):
        raise FileNotFoundError(f"couldnt find any hdf5 file in folder {folder}")
    for iF, filename in enumerate(sorted(glob.glob(os.path.join(folder, '*.hdf5')))):
        logger.info(f"reading {filename}")
        fin = h5py.File(filename, 'r')
        if 'deposited' in fin.attrs:
            deposited = fin.attrs['deposited']
            if prev_deposited is None:
                prev_deposited = deposited
            elif prev_deposited != deposited:
                raise AttributeError("The deposited parameter is not consistent among the input files!")

        if('trigger_names' in fin.attrs):
            trigger_names = fin.attrs['trigger_names']
        if(len(trigger_names) > 0):
            for iT, trigger_name in enumerate(trigger_names):
                trigger_names_dict[trigger_name] = iT
            break

    trigger_combinations['all_triggers'] = {'triggers': trigger_names}
    logger.info("Trigger names:", trigger_names)

    for iF, filename in enumerate(sorted(glob.glob(os.path.join(folder, '*.hdf5')))):
        fin = h5py.File(filename, 'r')
        out = {}
        E = fin.attrs['Emin']
        out['energy'] = E

        weights = np.array(fin['weights'])
        # triggered = np.array(fin['triggered'])
        triggered = get_triggered(fin)
        n_events = fin.attrs['n_events']
        if(trigger_names is None):
            trigger_names = fin.attrs['trigger_names']
            for iT, trigger_name in enumerate(trigger_names):
                trigger_names_dict[trigger_name] = iT
        else:
            if(np.any(trigger_names != fin.attrs['trigger_names'])):
                if(triggered.size == 0 and fin.attrs['trigger_names'].size == 0):
                    logger.warning("file {} has no triggering events. Using trigger names from another file".format(filename))
                else:
                    logger.warning("file {} has inconsistent trigger names: {}".format(filename, fin.attrs['trigger_names']))
                    raise

        # calculate effective
        rmin = fin.attrs['rmin']
        rmax = fin.attrs['rmax']
        thetamin = 0
        thetamax = np.pi
        phimin = 0
        phimax = 2 * np.pi
        if('thetamin' in fin.attrs):
            thetamin = fin.attrs['thetamin']
        if('thetamax' in fin.attrs):
            thetamax = fin.attrs['thetamax']
        if('phimin' in fin.attrs):
            fin.attrs['phimin']
        if('phimax' in fin.attrs):
            fin.attrs['phimax']
        dZ = fin.attrs['zmax'] - fin.attrs['zmin']
        area = np.pi * (rmax ** 2 - rmin ** 2)
        # The used area must be the projected area, perpendicular to the incoming
        # flux, which leaves us with the following correction. Remember that the
        # zenith bins must be small for the effective area to be correct.
        proj_area = area * 0.5 * (np.abs(np.cos(thetamin)) + np.abs(np.cos(thetamax)))
        V = area * dZ
        Vrms = fin.attrs['Vrms']

        # Solid angle needed for the effective volume calculations
        out['domega'] = np.abs(phimax - phimin) * np.abs(np.cos(thetamin) - np.cos(thetamax))
        out['thetamin'] = thetamin
        out['thetamax'] = thetamax
        out['deposited'] = deposited
        out['Aeffs'] = {}
        out['n_triggered_weighted'] = {}
        out['SNRs'] = {}

        if(triggered.size == 0):
            for iT, trigger_name in enumerate(trigger_names):
                out['Aeffs'][trigger_name] = [0, 0, 0]
            for trigger_name, values in iteritems(trigger_combinations):
                out['Aeffs'][trigger_name] = [0, 0, 0]
        else:
            for iT, trigger_name in enumerate(trigger_names):
                triggered = np.array(fin['multiple_triggers'][:, iT], dtype=np.bool)
                Aeff = proj_area * np.sum(weights[triggered]) / n_events
                Aeff_error = 0
                if(np.sum(weights[triggered]) > 0):
                    Aeff_error = Aeff / np.sum(weights[triggered]) ** 0.5
                out['Aeffs'][trigger_name] = [Aeff, Aeff_error, np.sum(weights[triggered])]

            for trigger_name, values in iteritems(trigger_combinations):
                indiv_triggers = values['triggers']
                triggered = np.zeros_like(fin['multiple_triggers'][:, 0], dtype=np.bool)
                if(isinstance(indiv_triggers, str)):
                    triggered = triggered | np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_triggers]], dtype=np.bool)
                else:
                    for indiv_trigger in indiv_triggers:
                        triggered = triggered | np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_trigger]], dtype=np.bool)
                if 'triggerAND' in values:
                    triggered = triggered & np.array(fin['multiple_triggers'][:, trigger_names_dict[values['triggerAND']]], dtype=np.bool)
                if 'notriggers' in values:
                    indiv_triggers = values['notriggers']
                    if(isinstance(indiv_triggers, str)):
                        triggered = triggered & ~np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_triggers]], dtype=np.bool)
                    else:
                        for indiv_trigger in indiv_triggers:
                            triggered = triggered & ~np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_trigger]], dtype=np.bool)
                if('min_sigma' in values.keys()):
                    if(isinstance(values['min_sigma'], list)):
                        if(trigger_name not in out['SNR']):
                            out['SNR'][trigger_name] = {}
                        masks = np.zeros_like(triggered)
                        for iS in range(len(values['min_sigma'])):
    #                         As = np.array(fin['maximum_amplitudes'])
                            As = np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)  # we use the this quantity because it is always computed before noise is added!
                            As_sorted = np.sort(As[:, values['channels'][iS]], axis=1)
                            # the smallest of the three largest amplitudes
                            max_amplitude = As_sorted[:, -values['n_channels'][iS]]
                            mask = np.sum(As[:, values['channels'][iS]] >= (values['min_sigma'][iS] * Vrms), axis=1) >= values['n_channels'][iS]
                            masks = masks | mask
                            out['SNR'][trigger_name][iS] = max_amplitude[mask] / Vrms
                        triggered = triggered & masks
                    else:
                        As = np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)  # we use the this quantity because it is always computed before noise is added!

                        As_sorted = np.sort(As[:, values['channels']], axis=1)
                        max_amplitude = As_sorted[:, -values['n_channels']]  # the smallest of the three largest amplitudes
                        mask = np.sum(As[:, values['channels']] >= (values['min_sigma'] * Vrms), axis=1) >= values['n_channels']

                        out['SNR'][trigger_name] = As_sorted[mask] / Vrms
                        triggered = triggered & mask
                if('ray_solution' in values.keys()):
                    As = np.array(fin['max_amp_ray_solution'])
                    max_amps = np.argmax(As[:, values['ray_channel']], axis=-1)
                    sol = np.array(fin['ray_tracing_solution_type'])
                    mask = np.array([sol[i, values['ray_channel'], max_amps[i]] == values['ray_solution'] for i in range(len(max_amps))], dtype=np.bool)
                    triggered = triggered & mask

                Aeff = proj_area * np.sum(weights[triggered]) / n_events

                if('efficiency' in values.keys()):
                    SNReff, eff = np.loadtxt("analysis_efficiency_{}.csv".format(values['efficiency']), delimiter=",", unpack=True)
                    get_eff = interpolate.interp1d(SNReff, eff, bounds_error=False, fill_value=(0, eff[-1]))
                    As = np.max(np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)[:, np.append(range(0, 8), range(12, 20))], axis=-1)  # we use the this quantity because it is always computed before noise is added!
                    if('efficiency_scale' in values.keys()):
                        As *= values['efficiency_scale']
                    e = get_eff(As / Vrms)
                    Aeff = proj_area * np.sum((weights * e)[triggered]) / n_events

                out['Aeffs'][trigger_name] = [Aeff, Aeff / np.sum(weights[triggered]) ** 0.5, np.sum(weights[triggered])]
        Aeff_output.append(out)

    return Aeff_output


def get_Veff_water_equivalent(Veff, density_medium=0.917 * units.g / units.cm ** 3, density_water=1 * units.g / units.cm ** 3):
    """
    convenience function to converte the effective volume of a medium with density `density_medium` to the
    water equivalent effective volume

    Parameters
    ----------
    Veff: float or array
        the effective volume
    dentity_medium: float (optional)
        the density of the medium of the Veff simulation (default deep ice)
    density water: float (optional)
        the density of water

    Returns: water equivalen effective volume
    """
    return Veff * density_medium / density_water


def get_Veff(folder, trigger_combinations={}, station=101, correct_zenith_sampling=False):
    """
    calculates the effective volume from NuRadioMC hdf5 files

    the effective volume is NOT normalized to a water equivalent. It is also NOT multiplied with the solid angle (typically 4pi).

    Parameters
    ----------
    folder: string
        folder conaining the hdf5 files, one per energy
    trigger_combinations: dict, optional
        keys are the names of triggers to calculate. Values are dicts again:
            * 'triggers': list of strings
                name of individual triggers that are combined with an OR
            the following additional options are optional
            * 'efficiency': string
                the signal efficiency vs. SNR (=Vmax/Vrms) to use. E.g. 'Chris'
            * 'efficiency_scale': float
                rescaling of the efficiency curve by SNR' = SNR * scale
            * 'n_reflections': int
                the number of bottom reflections of the ray tracing solution that likely triggered
                assuming that the solution with the shortest travel time caused the trigger, only considering channel 0

    station: int
        the station that should be considered
    correct_zenith_sampling: bool
        if True, correct a zenith sampling from np.sin(zenith) to an isotropic flux for a cylindrical geometry

    Returns
    ----------
    list of dictionary. Each file is one entry. The dictionary keys store all relevant properties
    """
    trigger_combinations = copy.copy(trigger_combinations)
    Veff_output = []
    trigger_names = None
    trigger_names_dict = {}
    prev_deposited = None
    deposited = False

    if(len(glob.glob(os.path.join(folder, '*.hdf5'))) == 0):
        raise FileNotFoundError(f"couldnt find any hdf5 file in folder {folder}")
    for iF, filename in enumerate(sorted(glob.glob(os.path.join(folder, '*.hdf5')))):
        logger.info(f"reading {filename}")
        fin = h5py.File(filename, 'r')
        if 'deposited' in fin.attrs:
            deposited = fin.attrs['deposited']
            if prev_deposited is None:
                prev_deposited = deposited
            elif prev_deposited != deposited:
                raise AttributeError("The deposited parameter is not consistent among the input files!")

        if('trigger_names' in fin.attrs):
            trigger_names = fin.attrs['trigger_names']
        if(len(trigger_names) > 0):
            for iT, trigger_name in enumerate(trigger_names):
                trigger_names_dict[trigger_name] = iT
            break

    trigger_combinations['all_triggers'] = {'triggers': trigger_names}
    logger.info("Trigger names:", trigger_names)
    for key in trigger_combinations:
        i = -1
        for value in trigger_combinations[key]['triggers']:
            i += 1
            if value not in trigger_names:
                logger.warning(f"trigger {value} not available, removing this trigger from the trigger combination {key}")
                trigger_combinations[key]['triggers'].pop(i)
                i -= 1

    for iF, filename in enumerate(sorted(glob.glob(os.path.join(folder, '*.hdf5')))):
        fin = h5py.File(filename, 'r')
        out = {}
        E = fin.attrs['Emin']
        if(fin.attrs['Emax'] != E):
            raise AttributeError("min and max energy do not match!")
        out['energy'] = E

        weights = np.array(fin['weights'])
        # triggered = np.array(fin['triggered'])
        triggered = get_triggered(fin)
        n_events = fin.attrs['n_events']
        if(trigger_names is None):
            trigger_names = fin.attrs['trigger_names']
            for iT, trigger_name in enumerate(trigger_names):
                trigger_names_dict[trigger_name] = iT
        else:
            if(np.any(trigger_names != fin.attrs['trigger_names'])):
                if(triggered.size == 0 and fin.attrs['trigger_names'].size == 0):
                    logger.warning("file {} has no triggering events. Using trigger names from another file".format(filename))
                else:
                    logger.error("file {} has inconsistent trigger names: {}".format(filename, fin.attrs['trigger_names']))
                    raise

        # calculate effective
        rmin = fin.attrs['rmin']
        rmax = fin.attrs['rmax']
        thetamin = 0
        thetamax = np.pi
        phimin = 0
        phimax = 2 * np.pi
        if('thetamin' in fin.attrs):
            thetamin = fin.attrs['thetamin']
        if('thetamax' in fin.attrs):
            thetamax = fin.attrs['thetamax']
        if('phimin' in fin.attrs):
            fin.attrs['phimin']
        if('phimax' in fin.attrs):
            fin.attrs['phimax']
        dZ = fin.attrs['zmax'] - fin.attrs['zmin']
        area = np.pi * (rmax ** 2 - rmin ** 2)
        V = area * dZ
        # calculate the average projected area in this zenith angle bin
        Aproj_avg = get_projected_area_cylinder_integral(thetamax, R=rmax, d=dZ) - get_projected_area_cylinder_integral(thetamin, R=rmax, d=dZ)
        # by not dividing by dCosTheta we automatically also integrate the solid angle into the weight.
        # Hence, simulations with zenith slices of different zenith angle coverage are possible.
        # Aproj_avg /= np.cos(thetamin) - np.cos(thetamax)
        out['Aproj'] = Aproj_avg
        Vrms = fin.attrs['Vrms']

        if(correct_zenith_sampling):
            if(len(weights) > 0):

                def get_weights(zeniths, thetamin, thetamax, R, d):
                    """
                    calculates a correction to the weight to go from a zenith distribution proportional from
                    theta ~ sin(theta) to an isotropic flux, i.e., the same number of events for the same 
                    projected area perpendicular to the incoming direction.  
                    
                    """
                    zeniths = np.array(zeniths)
                    yy = get_projected_area_cylinder(zeniths, R, d)
                    # calculate the average value of Aproj within the zenith band -> int(Aproc(theta) dcostheta)/int(1, dcostheta)
                    norm = get_projected_area_cylinder_integral(thetamax, R, d) - get_projected_area_cylinder_integral(thetamin, R, d)  # int(Aproc(theta) dcostheta)
                    norm /= (np.cos(thetamin) - np.cos(thetamax))  # int(1, dcostheta)
                    weights = yy / norm
                    logger.debug(f"{thetamin/units.deg:.0f} - {thetamax/units.deg:.0f}: average correction factor {weights.mean():.2f} max = {weights.max():.2f} min = {weights.min():.2f}")
                    return weights

                weights *= get_weights(fin['zeniths'], thetamin, thetamax, rmax, dZ)

        # Solid angle needed for the effective volume calculations
        out['domega'] = np.abs(phimax - phimin) * np.abs(np.cos(thetamin) - np.cos(thetamax))
        out['thetamin'] = thetamin
        out['thetamax'] = thetamax
        out['deposited'] = deposited
        out['Veffs'] = {}
        out['n_triggered_weighted'] = {}
        out['SNRs'] = {}

        if(triggered.size == 0):
            for iT, trigger_name in enumerate(trigger_names):
                out['Veffs'][trigger_name] = [0, 0, 0]
            for trigger_name, values in iteritems(trigger_combinations):
                out['Veffs'][trigger_name] = [0, 0, 0]
        else:
            for iT, trigger_name in enumerate(trigger_names):
                triggered = np.array(fin['multiple_triggers'][:, iT], dtype=np.bool)
                Veff = V * np.sum(weights[triggered]) / n_events
                Veff_error = 0
                if(np.sum(weights[triggered]) > 0):
                    Veff_error = Veff / np.sum(weights[triggered]) ** 0.5
                out['Veffs'][trigger_name] = [Veff, Veff_error, np.sum(weights[triggered])]

            for trigger_name, values in iteritems(trigger_combinations):
                indiv_triggers = values['triggers']
                triggered = np.zeros_like(fin['multiple_triggers'][:, 0], dtype=np.bool)
                if(isinstance(indiv_triggers, str)):
                    triggered = triggered | np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_triggers]], dtype=np.bool)
                else:
                    for indiv_trigger in indiv_triggers:
                        triggered = triggered | np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_trigger]], dtype=np.bool)
                if 'triggerAND' in values:
                    triggered = triggered & np.array(fin['multiple_triggers'][:, trigger_names_dict[values['triggerAND']]], dtype=np.bool)
                if 'notriggers' in values:
                    indiv_triggers = values['notriggers']
                    if(isinstance(indiv_triggers, str)):
                        triggered = triggered & ~np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_triggers]], dtype=np.bool)
                    else:
                        for indiv_trigger in indiv_triggers:
                            triggered = triggered & ~np.array(fin['multiple_triggers'][:, trigger_names_dict[indiv_trigger]], dtype=np.bool)
                if('min_sigma' in values.keys()):
                    if(isinstance(values['min_sigma'], list)):
                        if(trigger_name not in out['SNR']):
                            out['SNR'][trigger_name] = {}
                        masks = np.zeros_like(triggered)
                        for iS in range(len(values['min_sigma'])):
                            As = np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)  # we use the this quantity because it is always computed before noise is added!
                            As_sorted = np.sort(As[:, values['channels'][iS]], axis=1)
                            # the smallest of the three largest amplitudes
                            max_amplitude = As_sorted[:, -values['n_channels'][iS]]
                            mask = np.sum(As[:, values['channels'][iS]] >= (values['min_sigma'][iS] * Vrms), axis=1) >= values['n_channels'][iS]
                            masks = masks | mask
                            out['SNR'][trigger_name][iS] = max_amplitude[mask] / Vrms
                        triggered = triggered & masks
                    else:
                        As = np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)  # we use the this quantity because it is always computed before noise is added!

                        As_sorted = np.sort(As[:, values['channels']], axis=1)
                        max_amplitude = As_sorted[:, -values['n_channels']]  # the smallest of the three largest amplitudes

                        mask = np.sum(As[:, values['channels']] >= (values['min_sigma'] * Vrms), axis=1) >= values['n_channels']
                        out['SNR'][trigger_name] = As_sorted[mask] / Vrms
                        triggered = triggered & mask
                if('ray_solution' in values.keys()):
                    As = np.array(fin['max_amp_ray_solution'])
                    max_amps = np.argmax(As[:, values['ray_channel']], axis=-1)
                    sol = np.array(fin['ray_tracing_solution_type'])
                    mask = np.array([sol[i, values['ray_channel'], max_amps[i]] == values['ray_solution'] for i in range(len(max_amps))], dtype=np.bool)
                    triggered = triggered & mask

                if('n_reflections' in values.keys()):
                    if(np.sum(triggered)):
                        As = np.array(fin[f'station_{station:d}/max_amp_ray_solution'])
                        # find the ray tracing solution that produces the largest amplitude
                        max_amps = np.argmax(np.argmax(As[:, :], axis=-1), axis=-1)
                        # advanced indexing: selects the ray tracing solution per event with the highest amplitude
                        triggered = triggered & (np.array(fin[f'station_{station:d}/ray_tracing_reflection'])[..., max_amps, 0][:, 0] == values['n_reflections'])

                Veff = V * np.sum(weights[triggered]) / n_events

                if('efficiency' in values.keys()):
                    SNReff, eff = np.loadtxt("analysis_efficiency_{}.csv".format(values['efficiency']), delimiter=",", unpack=True)
                    get_eff = interpolate.interp1d(SNReff, eff, bounds_error=False, fill_value=(0, eff[-1]))
                    As = np.max(np.max(np.nan_to_num(fin['max_amp_ray_solution']), axis=-1)[:, np.append(range(0, 8), range(12, 20))], axis=-1)  # we use the this quantity because it is always computed before noise is added!
                    if('efficiency_scale' in values.keys()):
                        As *= values['efficiency_scale']
                    e = get_eff(As / Vrms)
                    Veff = V * np.sum((weights * e)[triggered]) / n_events

                Vefferror = 0
                if(np.sum(weights[triggered]) > 0):
                    Vefferror = Veff / np.sum(weights[triggered]) ** 0.5
                out['Veffs'][trigger_name] = [Veff, Vefferror, np.sum(weights[triggered])]
        Veff_output.append(out)

    return Veff_output


def get_Veff_array(data):
    """
    calculates a multi dimensional array of effective volume calculations for fast slicing

    the array dimensions are (energy, zenith bin, triggername, 3) where the
    last tuple is the effective volume, its uncertainty and the weighted sum of triggered events

    Parameters
    -----------
    data: dict
        the result of the `get_Veff` function

    Returns
    --------
     * (n_energy, n_zenith_bins, n_triggernames, 3) dimensional array of floats
     * array of unique energies
     * array of unique zenith bins
     * array of unique trigger names
     * array of weights for zenith averaging


    Examples
    ---------

    To plot the full sky effective volume for 'all_triggers' do

    ```
    output, uenergies, uzenith_bins, utrigger_names, zenith_weights = get_Veff_array(data)


    fig, ax = plt.subplots(1, 1)
    tname = "all_triggers"
    Veff = np.average(output[:,:,get_index(tname, utrigger_names),0], axis=1, weights=zenith_weights)
    Vefferror = Veff / np.sum(output[:,:,get_index(tname, utrigger_names),2], axis=1)**0.5
    ax.errorbar(uenergies/units.eV, Veff/units.km**3 * 4 * np.pi, yerr=Vefferror/units.km**3 * 4 * np.pi, fmt='-o', label=tname)

    ax.legend()
    ax.semilogy(True)
    ax.semilogx(True)
    fig.tight_layout()
    plt.show()
    ```


    To plot the effective volume for different declination bands do

    ```
    fig, ax = plt.subplots(1, 1)
    tname = "LPDA_2of4_100Hz"
    iZ = 9
    Veff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Veff[:,0]/units.km**3, yerr=Veff[:,1]/units.km**3,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")

    iZ = 8
    Veff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Veff[:,0]/units.km**3, yerr=Veff[:,1]/units.km**3,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")
    iZ = 7
    Veff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Veff[:,0]/units.km**3, yerr=Veff[:,1]/units.km**3,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")
    iZ = 10
    Veff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Veff[:,0]/units.km**3, yerr=Veff[:,1]/units.km**3,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")


    ax.legend()
    ax.semilogy(True)
    ax.semilogx(True)
    fig.tight_layout()
    plt.show()
    ```

    """
    energies = []
    zenith_bins = []
    trigger_names = []
    for d in data:
        energies.append(d['energy'])
        zenith_bins.append([d['thetamin'], d['thetamax']])
        for triggername in d['Veffs']:
            trigger_names.append(triggername)

    energies = np.array(energies)
    zenith_bins = np.array(zenith_bins)
    trigger_names = np.array(trigger_names)
    uenergies = np.unique(energies)
    uzenith_bins = np.unique(zenith_bins, axis=0)
    utrigger_names = np.unique(trigger_names)
    output = np.zeros((len(uenergies), len(uzenith_bins), len(utrigger_names), 3))
    weights = np.ones((len(uenergies), len(uzenith_bins)))
    logger.debug(f"unique energies {uenergies}")
    logger.debug(f"unique zenith angle bins {uzenith_bins/units.deg}")
    logger.debug(f"unique energies {utrigger_names}")

    for d in data:
        iE = np.squeeze(np.argwhere(d['energy'] == uenergies))
        iT = np.squeeze(np.argwhere([d['thetamin'], d['thetamax']] == uzenith_bins))[0][0]
        for triggername, Veff in d['Veffs'].items():
            iTrig = np.squeeze(np.argwhere(triggername == utrigger_names))
            output[iE, iT, iTrig] = Veff

    for d in data:
        iE = np.squeeze(np.argwhere(d['energy'] == uenergies))
        iT = np.squeeze(np.argwhere([d['thetamin'], d['thetamax']] == uzenith_bins))[0][0]
        if('Aproj' in d):
            weights[iE, iT] = d['Aproj']
    for iE in range(len(uenergies)):
        weights[iE] /= np.sum(weights[iE])

    return output, uenergies, uzenith_bins, utrigger_names, weights


def get_Aeff_array(data):
    """
    calculates a multi dimensional array of effective area calculations for fast slicing

    the array dimensions are (energy, zenith bin, triggername, 3) where the
    last tuple is the effective area, its uncertainty and the weighted sum of triggered events

    Parameters
    -----------
    data: dict
        the result of the `get_Aeff` function

    Returns
    --------
     * (n_energy, n_zenith_bins, n_triggernames, 3) dimensional array of floats
     * array of unique energies
     * array of unique zenith bins
     * array of unique trigger names


    Examples
    ---------

    To plot the full sky effective volume for 'all_triggers' do

    ```
    output, uenergies, uzenith_bins, utrigger_names = get_Aeff_array(data)


    fig, ax = plt.subplots(1, 1)
    tname = "all_triggers"
    Aeff = np.average(output[:,:,get_index(tname, utrigger_names),0], axis=1)
    Aefferror = Aeff / np.sum(output[:,:,get_index(tname, utrigger_names),2], axis=1)**0.5
    ax.errorbar(uenergies/units.eV, Aeff/units.km**3 * 4 * np.pi, yerr=Aefferror/units.km**3 * 4 * np.pi, fmt='-o', label=tname)

    ax.legend()
    ax.semilogy(True)
    ax.semilogx(True)
    fig.tight_layout()
    plt.show()
    ```


    To plot the effective area for different declination bands do

    ```
    fig, ax = plt.subplots(1, 1)
    tname = "LPDA_2of4_100Hz"
    iZ = 9
    Aeff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Aeff[:,0]/units.km**2, yerr=Aeff[:,1]/units.km**2,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")

    iZ = 8
    Aeff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Aeff[:,0]/units.km**2, yerr=Aeff[:,1]/units.km**2,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")
    iZ = 7
    Aeff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Aeff[:,0]/units.km**2, yerr=Aeff[:,1]/units.km**2,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")
    iZ = 10
    Aeff = output[:,iZ,get_index(tname, utrigger_names)]
    ax.errorbar(uenergies/units.eV, Aeff[:,0]/units.km**2, yerr=Aeff[:,1]/units.km**2,
                label=f"zenith bin {uzenith_bins[iZ][0]/units.deg:.0f} - {uzenith_bins[iZ][1]/units.deg:.0f}")


    ax.legend()
    ax.semilogy(True)
    ax.semilogx(True)
    fig.tight_layout()
    plt.show()
    ```

    """
    energies = []
    zenith_bins = []
    trigger_names = []
    for d in data:
        energies.append(d['energy'])
        zenith_bins.append([d['thetamin'], d['thetamax']])
        for triggername in d['Aeffs']:
            trigger_names.append(triggername)

    energies = np.array(energies)
    zenith_bins = np.array(zenith_bins)
    trigger_names = np.array(trigger_names)
    uenergies = np.unique(energies)
    uzenith_bins = np.unique(zenith_bins, axis=0)
    utrigger_names = np.unique(trigger_names)
    output = np.zeros((len(uenergies), len(uzenith_bins), len(utrigger_names), 3))
    logger.debug(f"unique energies {uenergies}")
    logger.debug(f"unique zenith angle bins {uzenith_bins/units.deg}")
    logger.debug(f"unique energies {utrigger_names}")

    for d in data:
        iE = np.squeeze(np.argwhere(d['energy'] == uenergies))
        iT = np.squeeze(np.argwhere([d['thetamin'], d['thetamax']] == uzenith_bins))[0][0]
        for triggername, Aeff in d['Aeffs'].items():
            iTrig = np.squeeze(np.argwhere(triggername == utrigger_names))
            output[iE, iT, iTrig] = Aeff

    return output, uenergies, uzenith_bins, utrigger_names


def get_index(value, array):
    return np.squeeze(np.argwhere(value == array))


def export(filename, data, trigger_names=None, export_format='yaml'):
    """
    export effective volumes (or effective areas) into a human readable JSON or YAML file

    Parameters
    ----------
    filename: string
        the output filename of the JSON file
    data: array
        the output of the `getVeff` function
    trigger_names: list of strings (optional, default None)
        save only specific trigger names, if None all triggers are exported
    export_format: string (default "yaml")
        specify output format, choose
        * "yaml"
        * "json"
    """
    output = []
    for i in range(len(data)):
        tmp = {}
        for key in data[i]:
            if (key not in  ['Veffs', 'Aeffs']):
                if isinstance(data[i][key], np.generic):
                    tmp[key] = data[i][key].item()
                else:
                    tmp[key] = data[i][key]
        for key in ["Veffs", "Aeffs"]:
            if(key in data[i]):
                tmp[key] = {}
                for trigger_name in data[i][key]:
                    if(trigger_names is None or trigger_name in trigger_names):
                        logger.info(trigger_name)
                        tmp[key][trigger_name] = []
                        for value in data[i][key][trigger_name]:
                            tmp[key][trigger_name].append(float(value))
        output.append(tmp)

    with open(filename, 'w') as fout:
        if(export_format == 'yaml'):
            import yaml
            yaml.dump(output, fout)
        elif(export_format == 'json'):
            json.dump(output, fout, sort_keys=True, indent=4)
